from typing import List

NODE_TYPE_TYPE = 1
NODE_TYPE_NAT_CONST = 2
NODE_TYPE_VAR_TYPE = 3
NODE_TYPE_VAR_NUM = 4
NODE_TYPE_ARRAY = 5

ID_VAR_NUM = 0x70659EFF
ID_VAR_TYPE = 0x2CECF817
ID_INT = 0xA8509BDA
ID_LONG = 0x22076CBA
ID_DOUBLE = 0x2210C154
ID_STRING = 0xB5286E24
ID_VECTOR = 0x1CB5C415
ID_DICTIONARY = 0x1F4C618F
ID_MAYBE_TRUE = 0x3F9C8EF8
ID_MAYBE_FALSE = 0x27930A7B
ID_BOOL_FALSE = 0xBC799737
ID_BOOL_TRUE = 0x997275B5

FLAG_OPT_VAR = 1 << 17
FLAG_EXCL = 1 << 18
FLAG_NOVAR = 1 << 21
FLAG_DEFAULT_CONSTRUCTOR = 1 << 25
FLAG_BARE = 1 << 0
FLAG_COMPLEX = 1 << 1
FLAGS_MASK = (1 << 16) - 1


class TlBase:
    def __repr__(self):
        return f'<{__name__}.{type(self).__name__}> {vars(self)}'

    def __str__(self):
        return repr(self)


class TlTree(TlBase):
    def __init__(
        self,
        flags: int,  # orig int32_t
    ):
        self.flags = flags

    def get_type(self):
        raise NotImplementedError


class TlTreeType(TlTree):
    def __init__(
        self,
        flags: int,
        type_: 'TlType',
        child_count: int,  # unused count of items to create list
    ):
        super().__init__(flags)
        self.type = type_
        self.children = []

    def get_type(self):
        return NODE_TYPE_TYPE


class TlTreeNatConst(TlTree):
    def __init__(
        self,
        flags: int,
        num: int,
    ):
        super().__init__(flags)
        self.num = num

    def get_type(self):
        return NODE_TYPE_NAT_CONST


class TlTreeVarType(TlTree):
    def __init__(
        self,
        flags: int,
        var_num: int,
    ):
        super().__init__(flags)
        self.var_num = var_num

    def get_type(self):
        return NODE_TYPE_VAR_TYPE


class TlTreeVarNum(TlTree):
    def __init__(
        self,
        flags: int,
        var_num: int,
        diff: int,
    ):
        super().__init__(flags)
        self.var_num = var_num
        self.diff = diff

    def get_type(self):
        return NODE_TYPE_VAR_NUM


class TlTreeArray(TlTree):
    def __init__(
        self,
        flags: int,
        multiplicity: TlTree,
        a: List['Arg'],
    ):
        super().__init__(flags)
        self.multiplicity = multiplicity
        self.args = a

    def get_type(self):
        return NODE_TYPE_ARRAY


class Arg(TlBase):
    def __init__(
        self,
        name: str = None,
        flags: int = None,  # orig int32_t
        var_num: int = None,
        exist_var_num: int = None,
        exist_var_bit: int = None,
        type: 'TlType' = None,
    ):
        self.name = name
        self.flags = flags
        self.var_num = var_num
        self.exist_var_num = exist_var_num
        self.exist_var_bit = exist_var_bit
        self.type = type


class TlCombinator(TlBase):
    def __init__(
        self,
        id_: int = None,  # orig int32_t
        name: str = None,
        var_count: int = None,
        type_id: int = None,  # orig int32_t
        args: List['Arg'] = None,
        result: 'TlTree' = None,
    ):
        self.id = id_
        self.name = name
        self.var_count = var_count
        self.type_id = type_id
        self.args = args
        self.result = result


class TlType(TlBase):
    def __init__(
        self,
        id_: int = None,  # orig int32_t
        name: str = None,
        arity: int = None,
        flags: int = None,  # orig int32_t
        simple_constructors: int = None,
        constructors_num: int = None,  # orig size_t
        constructors: List['TlCombinator'] = None,
    ):
        self.id = id_
        self.name = name
        self.arity = arity
        self.flags = flags
        self.simple_constructors = simple_constructors
        self.constructors_num = constructors_num
        self.constructors = constructors

    def add_constructor(self, new_constructor: 'TlCombinator') -> None:
        self.constructors.append(new_constructor)



import struct

format_int32 = 'I'
format_int64 = 'Q'
format_unsigned_char = 'B'

size_of_int32 = struct.calcsize(format_int32)
size_of_int64 = struct.calcsize(format_int64)
size_of_unsigned_char = struct.calcsize(format_unsigned_char)


class TlSimpleParser:
    def __init__(self, data: bytes):
        self.data = data
        self.data_begin = data
        self.data_len = len(data)  # size_t
        self.error = None  # str
        self.error_pos = None  # size_t

    def set_error(self, error_message: str) -> None:
        if self.error is not None:
            assert error_message is not None
            self.error = error_message
            self.error_pos = len(self.data_begin) - len(self.data)
            self.data = b'\x00\x00\x00\x00\x00\x00\x00\x00'
            self.data_len = 0
        else:
            self.data = b'\x00\x00\x00\x00\x00\x00\x00\x00'
            assert self.data_len == 0

    def check_len(self, len_: int) -> None:  # orig size_t
        if self.data_len < len_:
            self.set_error('Not enough data to read')
        else:
            self.data_len -= len_

    def get_error(self) -> str:
        return self.error

    def get_error_pos(self) -> int:  # orig size_t
        return self.error_pos

    def fetch_int(self) -> int:  # orig int32_t
        self.check_len(size_of_int32)
        result = struct.unpack(format_int32, self.data[:size_of_int32])[0]
        self.data = self.data[size_of_int32:]
        return result

    def fetch_long(self) -> int:  # orig int64_t
        self.check_len(size_of_int64)
        result = struct.unpack(format_int64, self.data[:size_of_int64])[0]
        self.data = self.data[size_of_int64:]
        return result

    def unpack_uchar(self, offset=0):
        return struct.unpack(format_unsigned_char, self.data[offset:size_of_unsigned_char])[0]

    def fetch_string(self) -> str:
        self.check_len(4)
        result_len = self.unpack_uchar()
        if result_len < 254:
            self.check_len((result_len >> 2) * 4)

            result = struct.unpack('c' * result_len, self.data[1 : result_len + 1])
            result_decoded = [c.decode('utf-8') for c in result]

            self.data = self.data[((result_len >> 2) + 1) * 4 :]
            return ''.join(result_decoded)

        if result_len == 254:
            result_len = self.unpack_uchar(1) + (self.unpack_uchar(2) << 8) + (self.unpack_uchar(3) << 16)
            self.check_len(((result_len + 3) >> 2) * 4)

            result = struct.unpack('c' * result_len, self.data[4 : result_len + 4])
            result_decoded = [c.decode('utf-8') for c in result]

            self.data = self.data[((result_len + 7) >> 2) * 4 :]
            return ''.join(result_decoded)

        self.set_error('Can\'t fetch string, 255 found')
        return ''

    def fetch_end(self) -> None:
        if self.data_len:
            self.set_error('Too much data to fetch')



import struct

from .tl_config import TlConfig
from .tl_config_parser import TlConfigParser


def read_tl_config(data: bytes) -> TlConfig:
    if not data:
        raise RuntimeError(f'Config data is empty')
    if len(data) % struct.calcsize('i') != 0:
        raise RuntimeError(f'Config size = {len(data)} is not multiple of {struct.calcsize("i")}')

    parser = TlConfigParser(data)
    return parser.parse_config()


def read_tl_config_from_file(file_name: str) -> TlConfig:
    with open(file_name, 'rb') as config:
        return read_tl_config(config.read())


__all__ = ['read_tl_config', 'read_tl_config_from_file']



from typing import List, Dict, Union

from .tl_core import TlBase, TlType, TlCombinator


class TlConfig(TlBase):
    def __init__(
        self,
    ):
        self.types: List['TlType'] = []
        self.id_to_type: Dict[int, 'TlType'] = {}  # orig int32_t
        self.name_to_type: Dict[str, 'TlType'] = {}
        self.functions: List['TlCombinator'] = []
        self.id_to_function: Dict[int, 'TlCombinator'] = {}  # orig int32_t
        self.name_to_function: Dict[str, 'TlCombinator'] = {}

    def add_type(self, type_: 'TlType') -> None:
        self.types.append(type_)
        self.id_to_type[type_.id] = type_
        self.name_to_type[type_.name] = type_

    def get_type(self, type_id_or_name: Union[int, str]) -> 'TlType':  # orig int32_t
        if isinstance(type_id_or_name, int):
            return self.id_to_type[type_id_or_name]
        else:
            return self.name_to_type[type_id_or_name]

    def add_function(self, function: 'TlCombinator') -> None:
        self.functions.append(function)
        self.id_to_function[function.id] = function
        self.name_to_function[function.name] = function

    def get_function(self, function_id_or_name: Union[int, str]) -> 'TlCombinator':  # orig int32_t
        if isinstance(function_id_or_name, int):
            return self.id_to_function[function_id_or_name]
        else:
            return self.name_to_function[function_id_or_name]

    def get_type_count(self) -> int:  # orig size_t
        return len(self.types)

    def get_type_by_num(self, num: int) -> 'TlType':  # orig size_t
        return self.types[num]

    def get_function_count(self) -> int:  # orig size_t
        return len(self.functions)

    def get_function_by_num(self, num: int) -> 'TlCombinator':  # orig size_t
        return self.functions[num]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['tlo']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'tlo',
    'version': '1.0.0',
    'description': 'Reader of binary serialized Type Language Schema',
    'long_description': '## Type Language Object\n\n> Reader of [binary serialized](https://core.telegram.org/mtproto/serialize) Type Language Schema\n\n### Example of result in the end of reading TLO\n\n<center>\n    <img src="https://raw.githubusercontent.com/MarshalX/tlo/main/.github/resources/demo.gif" alt="demo">\n</center>\n\n### Declaimer\n\nThis code has not been tested sufficiently. \nIt\'s a rewritten version of original reader in C++.\nIf you are going to use this for code generation, \nplease do additional tests. \nRecheck my implementation for errors and so on.\n\n### Context\n\nThe [Type Language (TL)](https://core.telegram.org/mtproto/TL) was\ninvented many years ago. It was originally used in [VK](https://vk.com/),\nand now in [Telegram](https://telegram.org). \nThe creators of this language invented and \nwrote all the necessary tools to work with it.\nFor example, a [parser of the language](https://github.com/vysheng/tl-parser)\nand its [binary format](https://core.telegram.org/mtproto/serialize)\nfor serialization was developed.\n\n### What is this for?\n\nTo work with TL Schemes using OOP. To generate the client MTProto code using\nofficial TL parsers and binary formats.\n\nMany Open Source MTProto client use their own implementation of parsers, \nwhich are not ultimatum. They are hardcoded for their minimal task.\n\nHardcode is not the way of Telegram. Official Telegram\'s Open Source projects \ntake the right approach. So, for example, [tdlib](https://github.com/tdlib/td)\ngenerates several interfaces for different languages and this is how it looks:\n\nRaw TL Schema -> Tl Parser -> binary TL Object -> **TLO reader** -> code generator.\n\n| Step name | Description |\n| --------- | ----------- |\n| Raw TL Schema  | Can be founded [here](https://core.telegram.org/schema) and in official Telegram repositories of client ([tdesktop/Telegram/Resources/tl](https://github.com/telegramdesktop/tdesktop/tree/dev/Telegram/Resources/tl), [tdlib/generate/scheme](https://github.com/tdlib/td/tree/master/td/generate/scheme)).  |\n| Tl Parser | Official TL parser written in C++. Now it\'s a part of [tdlib/td/generate/tl-parser](https://github.com/tdlib/td/tree/master/td/generate/tl-parser). In the input it takes raw TL schema file. The output is TLO file. |\n| binary TL Object | The output of Tl Parser. |\n| **TLO reader** | **This repository contains implementation of it in Python and JavaScript.** Reader of binary file. Provide access to combinators, types, functions, arguments and so on via Object Oriented Programming. |\n| code generator | Any code generator. In [tdlib/td/generate](https://github.com/tdlib/td/tree/master/td/generate) there is generator for C++, JNI, .NET and JSON interfaces. |\n\n### Installing\n\n#### For Python\n```bash\npip install tlo\n```\n\n#### ~~For JavaScript~~ Work in progress\n```bash\nnpm install tlo\n```\n\n### Usage\n\nYou can find TLO files for tests [here](https://github.com/MarshalX/tlo/tree/main/tlo_for_tests).\n\n#### Python (3.6+)\n```python\nfrom tlo import read_tl_config_from_file, read_tl_config\n\n\n# use read_tl_config(data) to pass bytes directly\nconfig = read_tl_config_from_file(\'td_api.tlo\')\n```\n\n#### ~~JavaScript~~ Work in progress\n```javascript\nimport {read_tl_config_from_file, read_tl_config} from \'tlo\';\n\n\n// use read_tl_config(data) to pass bytes directly\nconst config = read_tl_config_from_file(\'td_api.tlo\')\n```\n\n### Licence\n\nMIT License\n',
    'author': "Il'ya Semyonov",
    'author_email': 'ilya@marshal.dev',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



import logging

from .tl_core import *
from .tl_config import TlConfig
from .tl_simple_parser import TlSimpleParser

TLS_SCHEMA_V2 = 0x3A2F9BE2
TLS_SCHEMA_V3 = 0xE4A8604B
TLS_SCHEMA_V4 = 0x90AC88D7
TLS_TYPE = 0x12EB4386
TLS_COMBINATOR = 0x5C0A1ED5
TLS_COMBINATOR_LEFT_BUILTIN = 0xCD211F63
TLS_COMBINATOR_LEFT = 0x4C12C6D9
TLS_COMBINATOR_RIGHT_V2 = 0x2C064372
TLS_ARG_V2 = 0x29DFE61B

TLS_EXPR_NAT = 0xDCB49BD8
TLS_EXPR_TYPE = 0xECC9DA78

TLS_NAT_CONST_OLD = 0xDCB49BD8
TLS_NAT_CONST = 0x8CE940B1
TLS_NAT_VAR = 0x4E8A14F0
TLS_TYPE_VAR = 0x0142CEAE
TLS_ARRAY = 0xD9FB20DE
TLS_TYPE_EXPR = 0xC1863D08

logger = logging.getLogger(__name__)


class TlConfigParser(TlBase):
    def __init__(self, data: bytes):
        self.p = TlSimpleParser(data)
        self.schema_version = -1
        self.config = TlConfig()  # should be TlConfig

    def parse_config(self) -> 'TlConfig':
        self.schema_version = self.get_schema_version(self.try_parse_int())
        if self.schema_version < 2:
            raise RuntimeError(f'Unsupported tl-schema version {self.schema_version}')

        self.try_parse_int()  # date
        self.try_parse_int()  # version

        types_n = self.try_parse_int()
        constructors_total = 0
        for i in range(types_n):
            tl_type = self.read_type()
            self.config.add_type(tl_type)
            constructors_total += tl_type.constructors_num

        constructors_n = self.try_parse_int()
        assert constructors_n == constructors_total
        for i in range(constructors_n):
            tl_combinator = self.read_combinator()
            self.config.get_type(tl_combinator.type_id).add_constructor(tl_combinator)

        functions_n = self.try_parse_int()
        for i in range(functions_n):
            self.config.add_function(self.read_combinator())

        self.p.fetch_end()
        self.try_parse(0)

        return self.config

    @staticmethod
    def get_schema_version(version_id: int) -> int:
        if version_id == TLS_SCHEMA_V4:
            return 4
        elif version_id == TLS_SCHEMA_V3:
            return 3
        elif version_id == TLS_SCHEMA_V2:
            return 2

        return -1

    def read_type(self):
        t = self.try_parse_int()
        if t != TLS_TYPE:
            raise RuntimeError(f'Wrong tls_type magic {t}')

        tl_type = TlType()
        tl_type.id = self.try_parse_int()
        tl_type.name = self.try_parse_string()
        tl_type.constructors_num = self.try_parse_int()  # orig size_t
        tl_type.constructors = []

        tl_type.flags = self.try_parse_int()
        tl_type.flags &= ~(1 | 8 | 16 | 1024)
        if tl_type.flags != 0:
            logger.warning(f'Type {tl_type.name} has non-zero flags: {tl_type.flags}')

        tl_type.arity = self.try_parse_int()

        self.try_parse_long()  # unused

        return tl_type

    def read_combinator(self):
        t = self.try_parse_int()
        if t != TLS_COMBINATOR:
            raise RuntimeError(f'Wrong tls_combinator magic {t}')

        tl_combinator = TlCombinator()
        tl_combinator.id = self.try_parse_int()
        tl_combinator.name = self.try_parse_string()
        tl_combinator.type_id = self.try_parse_int()
        tl_combinator.var_count = 0

        left_type = self.try_parse_int()
        if left_type == TLS_COMBINATOR_LEFT:
            tl_combinator.args = self.read_args_list(tl_combinator)
        else:
            if left_type != TLS_COMBINATOR_LEFT_BUILTIN:
                raise RuntimeError(f'Wrong tls_combinator_left magic {left_type}')

        right_ver = self.try_parse_int()
        if right_ver != TLS_COMBINATOR_RIGHT_V2:
            raise RuntimeError(f'Wrong tls_combinator_right magic {right_ver}')

        tl_combinator.result = self.read_type_expr(tl_combinator)

        return tl_combinator

    def read_num_const(self) -> TlTree:
        num = self.try_parse_int()
        return TlTreeNatConst(FLAG_NOVAR, num)

    def read_num_var(self, tl_combinator: TlCombinator) -> TlTree:
        diff = self.try_parse_int()
        var_num = self.try_parse_int()

        if var_num >= tl_combinator.var_count:
            tl_combinator.var_count = var_num + 1

        return TlTreeVarNum(0, var_num, diff)

    def read_nat_expr(self, tl_combinator: TlCombinator) -> TlTree:
        tree_type = self.try_parse_int()
        if tree_type in (TLS_NAT_CONST_OLD, TLS_NAT_CONST):
            return self.read_num_const()
        elif tree_type == TLS_NAT_VAR:
            return self.read_num_var(tl_combinator)
        else:
            raise RuntimeError(f'tree_type = {tree_type}')

    def read_expr(self, tl_combinator: TlCombinator) -> TlTree:
        tree_type = self.try_parse_int()
        if tree_type == TLS_EXPR_NAT:
            return self.read_nat_expr(tl_combinator)
        elif tree_type == TLS_EXPR_TYPE:
            return self.read_type_expr(tl_combinator)
        else:
            raise RuntimeError(f'tree_type = {tree_type}')

    def read_args_list(self, tl_combinator: TlCombinator) -> List[Arg]:
        schema_flag_opt_field = 2 << int(self.schema_version >= 3)
        schema_flag_has_vars = schema_flag_opt_field ^ 6

        args_num = self.try_parse_int()
        args_list = []
        for i in range(args_num):
            arg = Arg()

            arg_v = self.try_parse_int()
            if arg_v != TLS_ARG_V2:
                raise RuntimeError(f'Wrong tls_arg magic {arg_v}')

            arg.name = self.try_parse_string()
            arg.flags = self.try_parse_int()

            is_optional = False
            if arg.flags & schema_flag_opt_field:
                arg.flags &= ~schema_flag_opt_field
                is_optional = True
            if arg.flags & schema_flag_has_vars:
                arg.flags &= ~schema_flag_has_vars
                arg.var_num = self.try_parse_int()
            else:
                arg.var_num = -1

            if arg.var_num >= tl_combinator.var_count:
                tl_combinator.var_count = arg.var_num + 1

            if is_optional:
                arg.exist_var_num = self.try_parse_int()
                arg.exist_var_bit = self.try_parse_int()
            else:
                arg.exist_var_num = -1
                arg.exist_var_bit = 0

            arg.type = self.read_type_expr(tl_combinator)
            if arg.type.flags & FLAG_NOVAR:
                arg.flags |= FLAG_NOVAR

            args_list.append(arg)

        return args_list

    def read_type_expr(self, tl_combinator: TlCombinator) -> TlTree:
        tree_type = self.try_parse_int()
        if tree_type == TLS_TYPE_VAR:
            return self.read_type_var(tl_combinator)
        elif tree_type == TLS_TYPE_EXPR:
            return self.read_type_tree(tl_combinator)
        elif tree_type == TLS_ARRAY:
            return self.read_array(tl_combinator)
        else:
            raise RuntimeError(f'tree_type = {tree_type}')

    def read_type_var(self, tl_combinator: TlCombinator) -> TlTree:
        var_num = self.try_parse_int()
        flags = self.try_parse_int()

        if var_num >= tl_combinator.var_count:
            tl_combinator.var_count = var_num + 1

        assert not (flags & (FLAG_NOVAR | FLAG_BARE))

        return TlTreeVarType(flags, var_num)

    def read_type_tree(self, tl_combinator: TlCombinator) -> TlTree:
        tl_type = self.config.get_type(self.try_parse_int())
        # there is assert not needed because we have KeyError exception
        flags = self.try_parse_int() | FLAG_NOVAR
        arity = self.try_parse_int()
        assert tl_type.arity == arity

        tl_tree_type = TlTreeType(flags, tl_type, arity)
        for i in range(arity):
            child = self.read_expr(tl_combinator)

            tl_tree_type.children.append(child)
            if not (child.flags & FLAG_NOVAR):
                tl_tree_type.flags &= ~FLAG_NOVAR

        return tl_tree_type

    def read_array(self, tl_combinator: TlCombinator) -> TlTree:
        flags = FLAG_NOVAR
        multiplicity = self.read_nat_expr(tl_combinator)

        tl_tree_array = TlTreeArray(flags, multiplicity, self.read_args_list(tl_combinator))

        for i in range(len(tl_tree_array.args)):
            if not (tl_tree_array.args[i].flags & FLAG_NOVAR):
                tl_tree_array.flags &= ~FLAG_NOVAR

        return tl_tree_array

    def try_parse(self, res):
        if self.p.get_error():
            raise RuntimeError(f'Wrong TL-scheme specified: {self.p.get_error()} at {self.p.get_error_pos()}')

        return res

    def try_parse_int(self) -> int:  # orig int32_t
        return self.try_parse(self.p.fetch_int())

    def try_parse_long(self) -> int:  # orig int64_t
        return self.try_parse(self.p.fetch_long())

    def try_parse_string(self) -> str:
        return self.try_parse(self.p.fetch_string())
